/*	
** Xmas (X memory access simulator) is a little memory (+cache) simulator.
** Copyright (C) 1997/98/99/00  <Paolo Dovera pd514314@silab.dsi.unimi.it>
**                              <Paolo Dovera pdovera@tiscalinet.it>

** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.

** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

** Send any bug reports or comments to pd514314@silab.dsi.unimi.it
**                                     pdovera@tiscalinet.it
*/

/*
** generals include
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>

/*
** include of application modules and images
*/
#include "opthandler.h"
#include "general_support.h"
#include "cache.h"
#include "memory.h"
#include "esp.h"
#include "global.h"

/* BEGIN LOCAL DEFINE */
#define MaxCountDemand 256 /* erase every MaxCountDemand access in cache */
#define ModuloZero 0       /* mod result equal to zero */

#define MinBitSizeCache 1
#define MaxBitSizeCache 0xb
#define DefaultCacheBitSize 5

/*
** define for WriteCacheMode Repeater
*/
#define WriteBackString    "WriteBack"
#define WriteThroughString "WriteThrough"

/*
** define for ReplaceMode Repeater
*/
#define SIDEString      "SIDE"
#define RANDOModeString "Random"
#define FIFOModeString  "FIFO"
#define LruModeString   "LeastRecentlyUsed"
#define LuModeString    "LeastUsed"
/* END LOCAL DEFINE */

/* BEGIN IMPORTED GLOABAL VARIABLES */
/* END IMPORTED GLOABAL VARIABLES */

/* BEGIN EXPORTABLE GLOBAL VARIABLES*/
/* END EXPORTABLE GLOBAL VARIABLES */

/* BEGIN STATIC FUNCTION (PRIVATE FUNCTION) */
/*
** Modify the demand field of an entry in cache
*/
static void cleardemand(unsigned int AddressCache);
static void incdemand(unsigned int AddressCache);
static void mru(unsigned int AddressCache, unsigned int startentry,
		unsigned int endentry);
static void incFIFO(unsigned int AddressCache, unsigned int startentry,
		 unsigned int endentry);

static int readincacheKeep(unsigned int AddressMem,unsigned int AddressCache,
			   int *addtoMiss);
static int writeincacheKeep(unsigned int AddressMem, unsigned int AddressCache,
			    int MemValue);
static int readincache(unsigned int AddressMem,unsigned int AddressCache,
		int *addtoMiss);
static int writeincache(unsigned int AddressMem, unsigned int AddressCache,
		 int MemValue);

static int LoadCacheControls(void);
static int LoadCacheValue(void);
/* END STATIC FUNCTION (PRIVATE FUNCTION) */

/* BEGIN STATIC GLOBAL VARIABLES (PRIVATE VARIABLES) */
struct EntryCache
{
  int Data;               /* word */
  unsigned int Tag;       /* address memory */
  unsigned int Valid : 1; /* entry validity bit 1 => valid data */
  unsigned int Dirty : 1; /* write back bit 1 => update */
  unsigned int Keep  : 1; /* 1 => no replace for this time */
  unsigned int Demand;    /* demand request */
};

static struct EntryCache* Cache=NULL; /* cache pointer */
static unsigned int MaxEntryCache;    /* max number of entry in a cache */
static unsigned int BitMemoryCache;
static unsigned int ModoScrittura; /* 0 => write throught 1 => write back */
static unsigned int ReplaceMode;   /*0 => least used 1 => least recently used*/
/* Miss:  counter of the read/write to memory */
/* Total: total counter of the request to a cache */
static unsigned int Miss,Total;
static unsigned int CountDemand; /* request counter */

static int (*ReadFunction)(unsigned int AddressMem,unsigned int AddressCache,
			   int *addtoMiss);
static int (*WriteFunction)(unsigned int AddressMem, unsigned int AddressCache,
			    int MemValue);
static int (*ReadFunctionNoStat)(unsigned int AddressMem,
				 unsigned int AddressCache, int *addtoMiss);
static int (*WriteFunctionNoStat)(unsigned int AddressMem,
				  unsigned int AddressCache, int MemValue);

static int ReshowCacheAscii=True;
/* END STATIC GLOBAL VARIABLES (PRIVATE VARIABLES) */

void inizializza_cache(void)
     /*
     ** Initialize the default state of the cache
     */
{
  /* default cache controls value */
  (void)setBitCache(DefaultCacheBitSize);
  ModoScrittura=WriteThrough;    /* default write mode */
  ReplaceMode=LruMode;           /* Least Recently Used */
  reset_statistic();
  LoadCacheControls(); /* read the cache controls */
  reset_cache_size();
  LoadCacheValue();
  ReadFunction=readincache;
  WriteFunction=writeincache;
  ReadFunctionNoStat=readincache;
  WriteFunctionNoStat=writeincache;
  srand((unsigned int)time(NULL));
}

void flushcache(void)
     /*
     ** Update the memory if needed
     */
{
  unsigned int entrycache;

  if (ModoScrittura == WriteBack)
    for (entrycache = 0; entrycache < MaxEntryCache; entrycache++) {
      if (Cache[entrycache].Dirty == True) { /* to memory */
	memwrite(Cache[entrycache].Tag,Cache[entrycache].Data);
	Cache[entrycache].Dirty=False;
      }
    }
}

void invalidcache(void)
     /*
     ** Update the memory if needed (WriteBack), invalid the cache, clean up
     ** the request counter and data
     */
{
  flushcache();
  memset(Cache,0,sizeof(struct EntryCache)*MaxEntryCache);
}

void reset_cache_size(void)
     /*
     ** Update the state of the cache for the new dimension
     */
{
  free(Cache);
  MaxEntryCache=(unsigned int)esp(2,BitMemoryCache);
  Cache=(struct EntryCache *)malloc(MaxEntryCache*sizeof(struct EntryCache));
  memset(Cache,0,sizeof(struct EntryCache)*MaxEntryCache);
}

static void cleardemand(unsigned int AddressCache)
     /*
     ** Clean up the requests of the entry
     */
{
  if (AddressCache < MaxEntryCache)
    Cache[AddressCache].Demand=0;
  else {
    printf("ERROR: Exceeded cache address on cleardemand %d > MaxEntryCache = %d\n",(int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

static void incdemand(unsigned int AddressCache)
     /*
     ** Go down the counter, every MaxCountDemand accessess clean up all
     ** the request for avoid particular problem
     */
{
  unsigned int entrycache;

  if (AddressCache < MaxEntryCache) {
      CountDemand++;
      if ((CountDemand % MaxCountDemand) == ModuloZero)
	for (entrycache = 0; entrycache < MaxEntryCache; entrycache++)
	  Cache[entrycache].Demand=0;  /* clean up the request */
      Cache[AddressCache].Demand++;
  }
  else {
    printf("Possible Error: Exceeded cache address on incdemand %d > MaxEntryCache = %d\n",(int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

static void mru(unsigned int AddressCache, unsigned int startentry,
		unsigned int endentry)
     /*
     ** Go up the time counter of the others entry and clean up the counter
     ** of the entry AddressCache
     */
{
  unsigned int entrycache;

   if (AddressCache < MaxEntryCache) {
     for (entrycache = startentry; entrycache < endentry; entrycache++)
       Cache[entrycache].Demand++;  /* get up the others entry */
     Cache[AddressCache].Demand=0;  /* entry requested */
   }
   else {
     printf("ERROR: Exceeded cache address on mru %d > MaxEntryCache = %d\n",(int)AddressCache,MaxEntryCache-1);
     exit(EXIT_FAILURE);
   }
}

unsigned int replace(unsigned int startentry, unsigned int endentry)
     /*
     ** Read the less used entry of cache for the replace mode
     */
{
  unsigned int entrycache,entryposition;
  unsigned int LeastUsed=0;

  entryposition=startentry;
  switch (ReplaceMode) {
  case LruMode:
    for (entrycache = startentry; entrycache < endentry; entrycache++) {
      /* cache scan */
      if (Cache[entrycache].Demand > LeastUsed) {
	entryposition=entrycache;          /* less used entry */
	LeastUsed=Cache[entrycache].Demand;  /* demand(i); */
      }
    }
    break;
  case LuMode:
    LeastUsed=UINT_MAX; /* upper bound for demand value: 2^32 */
    for (entrycache = startentry; entrycache < endentry; entrycache++) {
      /* cache scan */
      if (Cache[entrycache].Demand < LeastUsed) {
	entryposition=entrycache;          /* less used entry */
	LeastUsed=Cache[entrycache].Demand;  /* demand(i); */
      }
    }
    break;
  case FIFOMode:
    entryposition=Cache[startentry].Demand+startentry;
    break;
  case RANDOMode:
    entryposition=startentry+(int) (1.0*(endentry-startentry)*rand()/(RAND_MAX+1.0));
  }
  return(entryposition);
}

static void incFIFO(unsigned int AddressCache, unsigned int startentry,
		 unsigned int endentry)
     /*
     ** Go up the FIFO replacement position
     */
{
  Cache[startentry].Demand=(++Cache[startentry].Demand) % (endentry-startentry);
}

void Replacement(unsigned int AddressCache, unsigned int startentry,
		 unsigned int endentry)
     /*
     ** Update the replacement counter
     */
{
  switch (ReplaceMode) {
  case LruMode:
    mru(AddressCache,startentry,endentry);
    break;
  case LuMode:
    cleardemand(AddressCache);
    incdemand(AddressCache);
    break;
  case FIFOMode:
    incFIFO(AddressCache,startentry,endentry);
    break;
  }
}

int cacheWriteNoStat(unsigned int AddressMem, unsigned int AddressCache,
		int MemValue)
     /*
     ** Dummy function for write in cache without statistics update
     */
{
  return(WriteFunctionNoStat(AddressMem,AddressCache,MemValue));
}

void cachewrite(unsigned int AddressMem, unsigned int AddressCache,
		int MemValue)
     /*
     ** ci: AddressMem definito, AddressCache < MaxEntryCache, Value definito
     ** cu: write strategy: write back o write through
     */
{
  Miss+=WriteFunction(AddressMem,AddressCache,MemValue);
  Total++;
}

static int writeincacheKeep(unsigned int AddressMem, unsigned int AddressCache,
			    int MemValue)
     /*
     ** ci: AddressMem definito, AddressCache < MaxEntryCache,
     **     MemValue definito
     ** cu: write in cache without update the statistics, WITH KEEP FLAG ON
     */
{
  int addtoMiss=0;

  if (AddressCache < MaxEntryCache) {
    /* the cache address is valid */
    switch (ModoScrittura)
      {
      case WriteThrough :
	memwrite(AddressMem,MemValue);
	addtoMiss=1;
	if (Cache[AddressCache].Tag == AddressMem) {
	  Cache[AddressCache].Data=MemValue;
	  Cache[AddressCache].Keep=1;
	}
	else {
	  if (Cache[AddressCache].Keep == 0) {
	    Cache[AddressCache].Data=MemValue; /* write the data in cache */
	    Cache[AddressCache].Tag=AddressMem;
	    Cache[AddressCache].Valid=True;
	    Cache[AddressCache].Keep=1;
	  }
	  else
	    Cache[AddressCache].Keep=0;
	}
	return(addtoMiss);
	break;
      case WriteBack :
	if (Cache[AddressCache].Dirty == 1 &&
	    Cache[AddressCache].Valid == 1 &&
	    Cache[AddressCache].Tag == AddressMem) {
	  Cache[AddressCache].Data=MemValue;
	  Cache[AddressCache].Keep=1;
	}
	else {
	  addtoMiss=1;
	  if (Cache[AddressCache].Keep == 0) {
	    memwrite(Cache[AddressCache].Tag,Cache[AddressCache].Data);
	    Cache[AddressCache].Data=MemValue; /* write the data in cache */
	    Cache[AddressCache].Tag=AddressMem;
	    Cache[AddressCache].Valid=True;
	    Cache[AddressCache].Keep=1;
	  }
	  else
	    memwrite(AddressMem,MemValue);
	}
	break;
      default : printf("Modalita'=%d\n",(int)ModoScrittura);exit(EXIT_FAILURE);
      }
  }
  else {
    printf("ERROR: Exceeded cache address on write %d > MaxEntryCache = %d\n",
	   (int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(addtoMiss);
}

static int writeincache(unsigned int AddressMem, unsigned int AddressCache,
		 int MemValue)
     /*
     ** ci: AddressMem definito, AddressCache < MaxEntryCache,
     **     MemValue definito
     ** cu: write in cache without update the statistics
     */
{
  int addtoMiss=0;

  if (AddressCache < MaxEntryCache) {
    /* the cache address is right */
    switch (ModoScrittura)
      {
      case WriteThrough : memwrite(AddressMem,MemValue); addtoMiss=1;
                            break;
      case WriteBack : if (Cache[AddressCache].Dirty == 1 &&
			   Cache[AddressCache].Valid == 1 &&
			   Cache[AddressCache].Tag != AddressMem) {
	               memwrite(Cache[AddressCache].Tag,Cache[AddressCache].Data);
                       addtoMiss=1;
                       }
                       Cache[AddressCache].Dirty=True;
                       break;
      default : printf("Modalita'=%d\n",(int)ModoScrittura);exit(EXIT_FAILURE);
      }
    Cache[AddressCache].Data=MemValue;       /* write the data in cache */
    Cache[AddressCache].Tag=AddressMem;
    Cache[AddressCache].Valid=True;
  }
  else {
    printf("ERROR: Exceeded cache address on write %d > MaxEntryCache = %d\n",
	   (int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(addtoMiss);
}

int cacheReadNoStat(unsigned int AddressMem, unsigned int AddressCache,
		    int *addtoMiss)
     /*
     ** Dummy function for read in cache without statistics update
     */
{
  return(ReadFunctionNoStat(AddressMem,AddressCache,addtoMiss));
}

int cacheread(unsigned int AddressMem,unsigned int AddressCache)
     /*
     ** ci: AddressMem definito, AddressCache < MaxEntryCache
     ** cu: read from cache and update the statistics
     */
{
  int cachedatum,addtoMiss=0;

  cachedatum=ReadFunction(AddressMem,AddressCache,&addtoMiss);
  Miss+=addtoMiss;
  Total++;
  return(cachedatum);
}

static int readincacheKeep(unsigned int AddressMem, unsigned int AddressCache,
			   int *addtoMiss)
     /*
     ** Return the cache data if it exist else read it from memory and write
     ** it in cache, NOT update the statistics
     ** WITH KEEP FLAG ON
     */
{
  *addtoMiss=0;
  if (AddressCache < MaxEntryCache) {
    /* the address cache exist */
    if (Cache[AddressCache].Valid == True) {
      if (AddressMem != Cache[AddressCache].Tag) { 
	/* the requested data isn't in cache */
	if (Cache[AddressCache].Dirty == True && ModoScrittura == WriteBack) {
	  /* write in memory if the entry was dirty */
	  memwrite(Cache[AddressCache].Tag,Cache[AddressCache].Data);
	  Cache[AddressCache].Dirty=False;
	}  
      }
      else {
	Cache[AddressCache].Keep=1;
	return(Cache[AddressCache].Data);
      }
    }
    *addtoMiss=1;
    if (Cache[AddressCache].Keep == 0) {
      Cache[AddressCache].Data=memread(AddressMem);
      Cache[AddressCache].Tag=AddressMem;
      Cache[AddressCache].Valid=True;
      Cache[AddressCache].Keep=1;
      /* now the data is certanly in cache (may was in central memory) */
      return(Cache[AddressCache].Data); /* return the data answered */
    }
    else {
      Cache[AddressCache].Keep=0;
      return(memread(AddressMem)); /* return the data answered without
				      cache update */
    }
  }
  else {
    printf("ERROR: Exceeded cache address on read %d > MaxEntryCache = %d\n",
	   (int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

static int readincache(unsigned int AddressMem,unsigned int AddressCache,int *addtoMiss)
     /*
     ** Return the cache data if it exist else read it from memory and write
     ** it in cache, NOT update the statistics
     */
{

  *addtoMiss=0;
  if (AddressCache < MaxEntryCache) {
    /* the address cache exist */
    if (Cache[AddressCache].Valid == True) {
      if (AddressMem != Cache[AddressCache].Tag) { 
	/* the requested data isn't in cache */
	if (Cache[AddressCache].Dirty == True && ModoScrittura == WriteBack) {
	  /* write in memory if the entry was dirty */
	  memwrite(Cache[AddressCache].Tag,Cache[AddressCache].Data);
	  Cache[AddressCache].Dirty=False;
	}  
      }
      else
	return(Cache[AddressCache].Data);
    }
    Cache[AddressCache].Data=memread(AddressMem);
    Cache[AddressCache].Tag=AddressMem;
    Cache[AddressCache].Valid=True;
    *addtoMiss=1;
    /* now the data is certanly in cache (may was in central memory) */
    return(Cache[AddressCache].Data); /* return the data answered */
  }
  else {
    printf("ERROR: Exceeded cache address on read %d > MaxEntryCache = %d\n",
	   (int)AddressCache,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

void reset_statistic(void)
     /*
     ** Clean up the statistics
     */
{
  Miss=0;
  Total=0;
}

void statistic(unsigned int *miss, unsigned int *total)
     /*
     ** Return the statistic of the cache
     */
{
  *miss=Miss;
  *total=Total;
}

int data(unsigned int Entry)
     /*
     ** ci: Entry < MaxEntryCache
     ** cu: return the data of the entry
     */
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Data);
  else {
    printf("ERROR: Exceeded cache address on data %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int setdata(unsigned int Entry, int value)
     /*
     ** Write the data field at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Data=value;
  else {
    printf("ERROR: Exceeded cache address on setdata %d > MaxEntryCache = %d\n"
	   ,(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int tag(unsigned int Entry)
     /*
     ** ci: Entry < MaxEntryCache
     ** cu: return the Tag value of the entry
     */
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Tag);
  else {
    printf("ERROR: Exceeded cache address on tag %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int settag(unsigned int Entry, unsigned int value)
     /*
     ** Write the tag field  at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Tag=value;
  else {
    printf("ERROR: Exceeded cache address on settag %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int valid(unsigned int Entry)
     /*
     ** ci: Entry < MaxEntryCache
     ** cu: return the Valid value of the entry
     */
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Valid);
  else {
    printf("ERROR: Exceeded cache address on valid %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int setvalid(unsigned int Entry, unsigned int value)
     /*
     ** Write the valid field  at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Valid=value;
  else {
    printf("ERROR: Exceeded cache address on setvalid %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int dirty(unsigned int Entry)
     /*
     ** ci: Entry < MaxEntryCache
     ** cu: return the Dirty value of the entry
     */
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Dirty);
  else {
    printf("ERROR: Exceeded cache address on dirty %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int setdirty(unsigned int Entry, unsigned int value)
     /*
     ** Write the dirty field at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Dirty=value;
  else {
    printf("ERROR: Exceeded cache address on setdirty %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int getKeep(unsigned int Entry)
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Keep);
  else {
    printf("ERROR: Exceeded cache address on keep %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int setKeep(unsigned int Entry, unsigned int value)
     /*
     ** Write the keep field at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Keep=value;
  else {
    printf("ERROR: Exceeded cache address on setKeep %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int demand(unsigned int Entry)
     /*
     ** ci: Entry < MaxEntryCache
     ** cu: return the Demand value of the entry
     */
{
  if (Entry < MaxEntryCache)
    return(Cache[Entry].Demand);
  else {
    printf("ERROR: Exceeded cache address on demand %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
}

int setdemand(unsigned int Entry, unsigned int value)
     /*
     ** Write the demand field  at position Entry with value
     */
{
  if (Entry < MaxEntryCache)
    Cache[Entry].Demand=value;
  else {
    printf("ERROR: Exceeded cache address on setdemand %d > MaxEntryCache = %d\n",(int)Entry,MaxEntryCache-1);
    exit(EXIT_FAILURE);
  }
  return(True);
}

unsigned int getBitMemoryCache(void)
     /*
     ** cu: return the bit of cache
     */
{
  return(BitMemoryCache);
}

int setBitCache(unsigned int newbitcache)
     /*
     ** ci: newbitcache > 0
     ** cu: set the new value of cache bit
     */
{
  if ((newbitcache >= MinBitSizeCache) && (newbitcache <= MaxBitSizeCache)) {
    BitMemoryCache=newbitcache;
    return(True);
  }
  else
    return(False);
}

char *getStringWritemode(void)
     /*
     ** Return the associated string
     */{
  switch (ModoScrittura) {
  case WriteThrough: return(WriteThroughString); break;
  case WriteBack:    return(WriteBackString);    break;
  default : return(NULL);
  }
}

unsigned int getModoScrittura(void)
     /*
     ** cu: return the state of ModoScrittura
     */
{
  return(ModoScrittura);
}

unsigned int setModoScrittura(unsigned int newModoScrittura)
     /*
     ** ci: newModoScrittura > 0
     ** cu: set the value of ModoScrittura 0 => write throught 1 => write back
     */
{
  if ((newModoScrittura >= 0) && (newModoScrittura <= 1)) {
    ModoScrittura=newModoScrittura;
    return(True);
  }
  else
    return(False);
}

char *getStringReplacement(void)
     /*
     ** Return the associated string
     */{
  switch (ReplaceMode) {
  case LruMode: return(LruModeString); break;
  case LuMode: return(LuModeString); break;
  case FIFOMode: return(FIFOModeString); break;
  case RANDOMode: return(RANDOModeString); break;
  default: return(NULL);
  }
}

unsigned int getReplaceMode(void)
     /*
     ** cu: return the state of ReplaceMode
     */
{
  return(ReplaceMode);
}

unsigned int setReplaceMode(unsigned int newReplaceMode)
     /*
     ** ci: newReplaceMode > 0
     ** cu: set the value of ReplaceMode
     **     0 => least used 1 => least recently used
     */
{
  if ((newReplaceMode >= 0) && (newReplaceMode <= 3)) {
    ReplaceMode=newReplaceMode;
    return(True);
  }
  else
    return(False);
}

void ActiveCacheKeepMode(void)
     /*
     ** Switch to rd/wr functions for keep flags
     */
{
  ReadFunction=readincacheKeep;
  WriteFunction=writeincacheKeep;
  ReadFunctionNoStat=readincacheKeep;
  WriteFunctionNoStat=writeincacheKeep;
}

void ActiveCacheNormalMode(void)
     /*
     ** Switch to rd/wr functions for normal use
     */
{
  ReadFunction=readincache;
  WriteFunction=writeincache;
  ReadFunctionNoStat=readincache;
  WriteFunctionNoStat=writeincache;
}

unsigned int getCacheAlphaNumericStatus(void)
     /*
     ** Return true if cache AlphaNumeric is active
     */
{
  return(ReshowCacheAscii);
}

void setCacheAlphaNumericStatus(unsigned int status)
     /*
     ** Change the alfanumeric display cache mode
     */
{
  ReshowCacheAscii=status;
}

/*
** LOAD/SAVE cache status code after this line
** PLEASE PAY ATTENTION
*/

static int LoadCacheControls(void)
     /*
     ** load the cache controls and statistics
     */
{
  FILE *fp;
  char bufferline[MediumBufferLength],XmasDir[XmasDirLength],filename[FileNameLength];
  unsigned int writemode,replacemode,size;
  unsigned int total,miss,AlphaNumeric;

  XmasDirStatus(XmasDir,sizeof(XmasDir));
  sprintf(filename,"%s/cache",XmasDir);
  if ((fp=fopen(filename,"r")) != NULL) {
    while (fgets(bufferline,sizeof(bufferline),fp) != NULL) {
      if (strncmp(bufferline,CacheWriteOptString,strlen(CacheWriteOptString)) == 0)
	if (sscanf(bufferline+strlen(CacheWriteOptString),"%x",&writemode) == 1)
	  setModoScrittura(writemode);
      if (strncmp(bufferline,CacheReplaceOptString,strlen(CacheReplaceOptString)) == 0)
        if (sscanf(bufferline+strlen(CacheReplaceOptString),"%x",&replacemode) == 1)
	  setReplaceMode(replacemode);
      if (strncmp(bufferline,CacheSizeOptString,strlen(CacheSizeOptString)) == 0)
	if (sscanf(bufferline+strlen(CacheSizeOptString),"%x",&size) == 1)
	  setBitCache(size);
      if (strncmp(bufferline,CacheTotalOptString,strlen(CacheTotalOptString)) == 0)
	if (sscanf(bufferline+strlen(CacheTotalOptString),"%d",&total) == 1)
	  Total=total;
      if (strncmp(bufferline,CacheMissOptString,strlen(CacheMissOptString)) == 0)
	if (sscanf(bufferline+strlen(CacheMissOptString),"%d",&miss) == 1)
	  Miss=miss;
      if (strncmp(bufferline,CacheAlphaNumericOptString,
		  strlen(CacheAlphaNumericOptString)) == 0)
	if (sscanf(bufferline+strlen(CacheAlphaNumericOptString),"%d",
		   &AlphaNumeric) == 1)
	  ReshowCacheAscii=AlphaNumeric;
	  }
    fclose(fp);
    return(True);
  }
  return(False);
}

static int LoadCacheValue(void)
     /*
     ** Load the cache entry if the option CacheEntrySaveOpt is True (T)
     */
{
  FILE *fp;
  char CachePrf[FileNameLength];
  char RetValue[MediumBufferLength];
  char XmasDir[XmasDirLength];
  char row[MediumBufferLength];
  unsigned int i,dummy;
  struct EntryCache *TmpCache;
  unsigned int cacheaddress,cachevalue,cachetag,cachevalid,cachedirty,cachekeep,cachedemand;

  if (getOptionFromName(CachePreferencesFile,CacheEntrySaveOpt,RetValue,sizeof(RetValue)) == True) {
    if (getBoolOptionsValue(RetValue)) {
      XmasDirStatus(XmasDir,sizeof(XmasDir));
      sprintf(CachePrf,"%s/%s",XmasDir,CacheSaveDataFile);
      if ((fp=fopen(CachePrf,"r")) != NULL) {
	TmpCache=(struct EntryCache *)malloc(MaxEntryCache*
					     sizeof(struct EntryCache));
	memset(TmpCache,0,MaxEntryCache*sizeof(struct EntryCache));
	while (feof(fp) != True) { /* scan all file */
	  fgets(row,sizeof(row),fp);
	  if (strncmp(row,CacheEntrySaveLdc,strlen(CacheEntrySaveLdc)) == 0) {
	    sscanf(row,CacheEntrySaveLdc"%x=%x %x %x %x %x %x",&cacheaddress,&cachevalue,&cachetag,&cachevalid,&cachedirty,&cachekeep,&cachedemand);
	    TmpCache[cacheaddress].Data=cachevalue;
	    TmpCache[cacheaddress].Tag=cachetag;
	    TmpCache[cacheaddress].Valid=cachevalid;
	    TmpCache[cacheaddress].Dirty=cachedirty;
	    TmpCache[cacheaddress].Keep=cachekeep;
	    TmpCache[cacheaddress].Demand=cachedemand;
	  }
	}
	/*
	** dummy is used to avoid boring warning at compile-time
	*/
	for (i=0; i<MaxEntryCache; i++) {
	  setdata(i,TmpCache[i].Data);
	  /*Cache[i].Data=TmpCache[i].Data;*/
	  settag(i,TmpCache[i].Tag);
	  /*Cache[i].Tag=TmpCache[i].Tag;*/
	  setvalid(i,(dummy=TmpCache[i].Valid));
	  /*Cache[i].Valid=TmpCache[i].Valid;*/
	  setdirty(i,(dummy=TmpCache[i].Dirty));
	  /*Cache[i].Dirty=TmpCache[i].Dirty;*/
	  setKeep(i,(dummy=TmpCache[i].Keep));
	  /*Cache[i].Keep=TmpCache[i].Keep;*/
	  setdemand(i,TmpCache[i].Demand);
	  /*Cache[i].Demand=TmpCache[i].Demand;*/
	}
	free(TmpCache);
	fclose(fp);
	return(True);
      }
    }
  }
  return(False);
}

int SaveCacheStatus(char *XmasDir)
     /*
     ** save the cache contols and value
     */
{
  FILE *fp;
  char filename[FileNameLength];
  char RetValue[MediumBufferLength];
  unsigned int i;

  sprintf(filename,"%s/%s",XmasDir,CacheSaveDataFile);
  if ((fp=fopen(filename,"w")) != NULL) {
    fprintf(fp,"%s 0x%x\n",CacheWriteOptString,ModoScrittura);
    fprintf(fp,"%s 0x%x\n",CacheReplaceOptString,ReplaceMode);
    fprintf(fp,"%s 0x%x\n",CacheSizeOptString,BitMemoryCache);
    fprintf(fp,"%s %d\n",CacheTotalOptString,Total);
    fprintf(fp,"%s %d\n",CacheMissOptString,Miss);
    fprintf(fp,"%s %d\n",CacheAlphaNumericOptString,ReshowCacheAscii);
    if (getOptionFromName(CachePreferencesFile,CacheEntrySaveOpt,RetValue,sizeof(RetValue)) == True) {
      if (getBoolOptionsValue(RetValue))
	for (i=0; i<MaxEntryCache; i++)
	  fprintf(fp,"%s%x=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",CacheEntrySaveLdc,i,data(i),tag(i),valid(i),dirty(i),getKeep(i),demand(i));
    }
    fclose(fp);
    return(True);
  }
    return(False);
}
/*
** LOAD/SAVE cache status code before this line
** PLEASE PAY ATTENTION
*/
